#!/usr/local/bin/python
# encoding: utf-8
"""
*The classifier object for Sherlock*

:Author:
    David Young

:Date Created:
    June 29, 2015

:Notes:
    - If you have any questions requiring this script/module please email me: d.r.young@qub.ac.uk
"""
################# GLOBAL IMPORTS ####################
import sys
import os
os.environ['TERM'] = 'vt100'
import readline
import glob
import pickle
import time
import MySQLdb as ms
from docopt import docopt
from dryxPython import logs as dl
from dryxPython import commonutils as dcu
from dryxPython import mysql as dms
from fundamentals import tools, times
from sherlock.update_ned_stream import update_ned_stream


class classifier():

    """
    *The classifier object for Sherlock*

    **Key Arguments:**
        - ``log`` -- logger
        - ``settings`` -- the settings dictionary
        - ``update`` -- update the transient database with crossmatch results (boolean)
    """
    # INITIALISATION

    def __init__(
            self,
            log,
            settings=False,
            update=False
    ):
        self.log = log
        log.debug("instansiating a new 'classifier' object")
        self.settings = settings
        self.update = update
        # xt-self-arg-tmpx

        # VARIABLE DATA ATRRIBUTES
        self.transientsMetadataList = []

        # INITIAL ACTIONS
        # SETUP DATABASE CONNECTIONS
        from sherlock import database
        db = database(
            log=self.log,
            settings=self.settings
        )
        self.transientsDbConn, self.cataloguesDbConn, self.pmDbConn = db.get()
        self.crossmatchTablename = self.settings[
            "database settings"]["transients"]["crossmatchTable"]

        return None

    # METHOD ATTRIBUTES

    def get(self):
        """
        *perform the classifications*
        """
        self.log.debug('starting the ``get`` method')

        self._create_crossmatch_table_if_not_existing()
        self._grab_column_name_map_from_database()

        # GRAB TRANSIENT METADATA BEFORE CLASSIFICATION
        self.transientsMetadataList = self._get_transient_metadata_from_sqlquery()

        updater = update_ned_stream(
            log=self.log,
            cataloguesDbConn=self.cataloguesDbConn,
            settings=self.settings,
            transientsMetadataList=self.transientsMetadataList
        ).get()
        del updater

        self._crossmatch_transients_against_catalogues()

        if self.update:
            self._update_transient_database()

        self.log.debug('completed the ``get`` method')
        return None

    def _get_transient_metadata_from_sqlquery(
            self):
        """ get transient metadata from a given workflow list in the transient database
        """
        self.log.debug(
            'starting the ``_get_transient_metadata_from_sqlquery`` method')

        sqlQuery = self.settings["database settings"][
            "transients"]["transient query"]
        transientsMetadataList = dms.execute_mysql_read_query(
            sqlQuery=sqlQuery,
            dbConn=self.transientsDbConn,
            log=self.log
        )

        self.log.debug(
            'completed the ``_get_transient_metadata_from_sqlquery`` method')
        return transientsMetadataList

    def _get_transient_metadata_from_database_list(
            self):
        """
        *get transient metadata from a given workflow list in the transient database*
        """
        self.log.debug(
            'starting the ``_get_transient_metadata_from_database_list`` method')

        sqlQuery = self.settings["database settings"][
            "transients"]["transient query"]
        self.transientsMetadataList = dms.execute_mysql_read_query(
            sqlQuery=sqlQuery,
            dbConn=self.transientsDbConn,
            log=self.log
        )

        self.log.debug(
            'completed the ``_get_transient_metadata_from_database_list`` method')
        return None

    def _crossmatch_transients_against_catalogues(
            self):
        """
        *crossmatch transients against catalogue*

        **Key Arguments:**
            # -

        **Return:**
            - None

        .. todo::

            - @review: when complete, clean _crossmatch_transients_against_catalogues method
            - @review: when complete add logging
        """
        self.log.debug(
            'starting the ``_crossmatch_transients_against_catalogues`` method')

        from sherlock import crossmatcher
        self.allClassifications = []

        cm = crossmatcher(
            log=self.log,
            dbConn=self.cataloguesDbConn,
            transients=self.transientsMetadataList,
            settings=self.settings,
            colMaps=self.colMaps
        )
        self.classifications = cm.get()
        del cm

        self.log.debug(
            'completed the ``_crossmatch_transients_against_catalogues`` method')
        return None

    # use the tab-trigger below for new method
    def _update_transient_database(
            self):
        """
        *update transient database*

        **Key Arguments:**
            # -

        **Return:**
            - None

        .. todo::

            - @review: when complete, clean _update_transient_database method
            - @review: when complete add logging
        """
        self.log.debug('starting the ``_update_transient_database`` method')

        crossmatchTablename = self.crossmatchTablename

        from datetime import datetime, date, time
        now = datetime.now()
        now = now.strftime("%Y-%m-%d %H:%M:%S")

        transientTable = self.settings["database settings"][
            "transients"]["transient table"]
        transientTableClassCol = self.settings["database settings"][
            "transients"]["transient classification column"]
        transientTableIdCol = self.settings["database settings"][
            "transients"]["transient primary id column"]

        for c in self.classifications:

            objectType = c["object_classification_new"]
            transientObjectId = c["id"]

            # DELETE PREVIOUS CROSSMATCHES
            try:
                thisId = int(transientObjectId)
            except:
                thisId = '"%(transientObjectId)s"' % locals()

            sqlQuery = u"""
                  delete from %(crossmatchTablename)s  where transient_object_id = %(thisId)s
            """ % locals()
            dms.execute_mysql_write_query(
                sqlQuery=sqlQuery,
                dbConn=self.transientsDbConn,
                log=self.log
            )

            # INSERT NEW CROSSMATCHES
            sqlQuery = u"""INSERT ignore into %(crossmatchTablename)s""" % locals(
            )
            sqlQuery += u""" (
                           transient_object_id,
                           catalogue_object_id,
                           catalogue_table_id,
                           catalogue_view_id,
                           catalogue_object_ra,
                           catalogue_object_dec,
                           catalogue_object_mag,
                           catalogue_object_filter,
                           original_search_radius_arcsec,
                           separation,
                           z,
                           scale,
                           distance,
                           distance_modulus,
                           date_added,
                           association_type,
                           physical_separation_kpc,
                           catalogue_object_type,
                           catalogue_object_subtype,
                           catalogue_table_name,
                           catalogue_view_name,
                           search_name,
                           major_axis_arcsec,
                           direct_distance,
                           direct_distance_scale,
                           direct_distance_modulus
                           )
                        values (
                           %s,
                           "%s",
                           %s,
                           %s,
                           %s,
                           %s,
                           %s,
                           %s,
                           %s,
                           %s,
                           %s,
                           %s,
                           %s,
                           %s,
                           %s,
                           %s,
                           %s,
                           %s,
                           %s,
                           %s,
                           %s,
                           %s,
                           %s,
                           %s,
                           %s,
                           %s)
                        """
            manyValueList = []
            for crossmatch in c["crossmatches"]:
                for k, v in crossmatch.iteritems():
                    if v == None:
                        crossmatch[k] = None
                if "physical_separation_kpc" not in crossmatch.keys():
                    crossmatch["physical_separation_kpc"] = None
                if "sourceFilter" not in crossmatch.keys():
                    crossmatch["sourceFilter"] = None
                if "sourceMagnitude" not in crossmatch.keys():
                    crossmatch["sourceMagnitude"] = None

                if crossmatch["sourceSubType"] and "null" not in str(crossmatch["sourceSubType"]):
                    crossmatch["sourceSubType"] = '"%s"' % (crossmatch[
                        "sourceSubType"],)
                else:
                    crossmatch["sourceSubType"] = None

                for k, v in crossmatch.iteritems():
                    if v == "null":
                        crossmatch[k] = None

                theseValues = (thisId, crossmatch["catalogueObjectId"], crossmatch["catalogueTableId"], crossmatch["catalogueViewId"], crossmatch["sourceRa"], crossmatch["sourceDec"], crossmatch["sourceMagnitude"], crossmatch["sourceFilter"],  crossmatch["originalSearchRadius"], crossmatch["separation"], crossmatch["z"], crossmatch["scale"], crossmatch["distance"], crossmatch[
                               "distanceModulus"], now, crossmatch["association_type"], crossmatch["physical_separation_kpc"], crossmatch["sourceType"], crossmatch["sourceSubType"], crossmatch["catalogueTableName"], crossmatch["catalogueViewName"], crossmatch["searchName"], crossmatch["xmmajoraxis"], crossmatch["xmdirectdistance"], crossmatch["xmdirectdistancescale"], crossmatch["xmdirectdistanceModulus"])
                manyValueList.append(theseValues)

            dms.execute_mysql_write_query(
                sqlQuery=sqlQuery,
                dbConn=self.transientsDbConn,
                log=self.log,
                manyValueList=manyValueList
            )

        for ob in self.transientsMetadataList:
            transId = ob["id"]
            if not isinstance(transId, int):
                transId = '"%(transId)s"' % locals()
            name = ob["name"]

            sqlQuery = u"""
                select id, separation, catalogue_view_name, association_type, physical_separation_kpc, major_axis_arcsec from %(crossmatchTablename)s  where transient_object_id = %(transId)s order by separation
            """ % locals()
            rows = dms.execute_mysql_read_query(
                sqlQuery=sqlQuery,
                dbConn=self.transientsDbConn,
                log=self.log
            )

            rankScores = []
            for row in rows:
                if row["separation"] < 2. or (row["physical_separation_kpc"] != None and row["physical_separation_kpc"] < 20. and row["association_type"] == "SN") or (row["major_axis_arcsec"] != "null" and row["association_type"] == "SN"):
                    # print row["separation"]
                    # print row["physical_separation_kpc"]
                    # print row["major_axis_arcsec"]
                    rankScore = 2. - \
                        self.colMaps[row["catalogue_view_name"]][
                            "object_type_accuracy"] * 0.1
                    # print rankScore
                else:
                    # print row["separation"]
                    # print row["physical_separation_kpc"]
                    # print row["major_axis_arcsec"]
                    rankScore = row["separation"] + 1. - \
                        self.colMaps[row["catalogue_view_name"]][
                            "object_type_accuracy"] * 0.1
                rankScores.append(rankScore)

            rank = 0
            theseValues = []
            for rs, row in sorted(zip(rankScores, rows)):
                rank += 1
                primaryId = row["id"]
                theseValues.append("(%(primaryId)s,%(rank)s)" % locals())

            theseValues = ",".join(theseValues)
            if len(theseValues):
                sqlQuery = u"""
                    INSERT INTO %(crossmatchTablename)s (id,rank) VALUES %(theseValues)s 
                    ON DUPLICATE KEY UPDATE rank=VALUES(rank);
                """ % locals()
                dms.execute_mysql_write_query(
                    sqlQuery=sqlQuery,
                    dbConn=self.transientsDbConn,
                    log=self.log,
                )

            sqlQuery = u"""
               select distinct association_type from (select association_type from %(crossmatchTablename)s  where transient_object_id = %(transId)s  order by rank) as alias;
            """ % locals()
            rows = dms.execute_mysql_read_query(
                sqlQuery=sqlQuery,
                dbConn=self.transientsDbConn,
                log=self.log
            )

            classification = ""
            for row in rows:
                classification += row["association_type"] + "/"
            classification = classification[:-1]

            if len(classification) == 0:
                classification = "ORPHAN"

            sqlQuery = u"""
                    update %(transientTable)s  set %(transientTableClassCol)s = "%(classification)s"
                        where %(transientTableIdCol)s  = %(transId)s
                """ % locals()

            print """%(name)s: %(classification)s """ % locals()

            dms.execute_mysql_write_query(
                sqlQuery=sqlQuery,
                dbConn=self.transientsDbConn,
                log=self.log
            )

        self.log.debug('completed the ``_update_transient_database`` method')
        return None

    def _grab_column_name_map_from_database(
            self):
        """
        *grab column name map from database*

        **Return:**
            - None
        """
        self.log.info(
            'starting the ``_grab_column_name_map_from_database`` method')

        # GRAB THE NAMES OF THE IMPORTANT COLUMNS FROM DATABASE
        sqlQuery = u"""
            select v.id as view_id, view_name, raColName, decColName, object_type, subTypeColName, objectNameColName, redshiftColName, distanceColName, semiMajorColName, semiMajorToArcsec, table_id, table_name, object_type_accuracy, filter1ColName, filterErr1ColName, filterName1ColName, filter2ColName, filterErr2ColName, filterName2ColName, filter3ColName, filterErr3ColName, filterName3ColName, filter4ColName, filterErr4ColName, filterName4ColName, filter5ColName, filterErr5ColName, filterName5ColName from tcs_helper_catalogue_views_info v, tcs_helper_catalogue_tables_info t where v.table_id = t.id
        """ % locals()
        rows = dms.execute_mysql_read_query(
            sqlQuery=sqlQuery,
            dbConn=self.cataloguesDbConn,
            log=self.log
        )
        self.colMaps = {}
        for row in rows:
            self.colMaps[row["view_name"]] = row

        self.log.info(
            'completed the ``_grab_column_name_map_from_database`` method')
        return None

    def _create_crossmatch_table_if_not_existing(
            self):
        """
        *create crossmatch table if it does not yet exist in the transients database*

        **Return:**
            - None
        """
        self.log.info(
            'starting the ``_create_crossmatch_table_if_not_existing`` method')

        crossmatchTablename = self.crossmatchTablename

        sqlQuery = u"""
            CREATE TABLE `%(crossmatchTablename)s` (
              `transient_object_id` varchar(50) NOT NULL DEFAULT '---',
              `catalogue_object_id` varchar(30) NOT NULL DEFAULT '---',
              `catalogue_table_id` smallint(5) unsigned NOT NULL DEFAULT '0',
              `separation` double DEFAULT NULL,
              `id` bigint(20) unsigned NOT NULL AUTO_INCREMENT,
              `z` double DEFAULT NULL,
              `scale` double DEFAULT NULL,
              `distance` double DEFAULT NULL,
              `distance_modulus` double DEFAULT NULL,
              `association_type` varchar(45) DEFAULT NULL,
              `date_added` datetime DEFAULT NULL,
              `physical_separation_kpc` double DEFAULT NULL,
              `catalogue_object_type` varchar(45) DEFAULT NULL,
              `catalogue_object_subtype` varchar(45) DEFAULT NULL,
              `association_rank` int(11) DEFAULT NULL,
              `catalogue_table_name` varchar(100) DEFAULT NULL,
              `catalogue_view_name` varchar(100) DEFAULT NULL,
              `rank` int(11) DEFAULT NULL,
              `search_name` varchar(100) DEFAULT NULL,
              `major_axis_arcsec` double DEFAULT NULL,
              `direct_distance` double DEFAULT NULL,
              `direct_distance_scale` double DEFAULT NULL,
              `direct_distance_modulus` double DEFAULT NULL,
              `catalogue_object_ra` double DEFAULT NULL,
              `catalogue_object_dec` double DEFAULT NULL,
              `original_search_radius_arcsec` double DEFAULT NULL,
              `catalogue_view_id` int(11) DEFAULT NULL,
              `catalogue_object_mag` float DEFAULT NULL,
              `catalogue_object_filter` varchar(10) DEFAULT NULL,
              PRIMARY KEY (`id`),
              UNIQUE KEY `transId_cata_obj_id` (`transient_object_id`,`catalogue_object_id`),
              UNIQUE KEY `catid_search_name` (`catalogue_object_id`,`catalogue_table_id`,`search_name`),
              KEY `key_transient_object_id` (`transient_object_id`),
              KEY `key_catalogue_object_id` (`catalogue_object_id`),
              KEY `idx_separation` (`separation`)
            ) ENGINE=MyISAM AUTO_INCREMENT=15168904 DEFAULT CHARSET=latin1 DELAY_KEY_WRITE=1;
        """ % locals()
        try:
            dms.execute_mysql_write_query(
                sqlQuery=sqlQuery,
                dbConn=self.transientsDbConn,
                log=self.log
            )
        except:
            pass

        self.log.info(
            'completed the ``_create_crossmatch_table_if_not_existing`` method')
        return None

    # use the tab-trigger below for new method
    # xt-class-method


if __name__ == '__main__':
    main()



#!/usr/local/bin/python
# encoding: utf-8
"""
*Build and import NED-D table in datasbe*

:Author:
    David Young

:Date Created:
    September 7, 2015
"""
################# GLOBAL IMPORTS ####################
import sys
import os
os.environ['TERM'] = 'vt100'
import readline
import glob
import pickle
import csv
import time
from docopt import docopt
from dryxPython import logs as dl
from dryxPython import commonutils as dcu
from dryxPython import mysql as dms
from dryxPython import astrotools as dat
from neddy import namesearch
from fundamentals import tools, times
from .._base_importer import _base_importer
from . import *


class ned_d(_base_importer):

    """
    *The worker class for the ned_d importer module*
    """

    def get(self):
        """
        *import the NED-D database into the database and do other cleanup tricks.*
        """
        self.log.info('starting the ``get`` method')

        self.databaseInsertbatchSize = 10000
        self.uniqueKeyList = ["galaxy_index_id", "dist_index_id"]
        self.generate_the_database_table()
        self.dictList = self.create_dictionary_of_nedd()
        self.add_data_to_database_table()
        self.clean_up_columns()
        self.create_the_master_view()
        self.get_metadata_for_galaxies()
        self.add_htmids_to_database_table()
        self._update_sdss_coverage()

        self.log.info('completed the ``get`` method')
        return ned_d

    def generate_the_database_table(
            self):
        """
        *generate the NED-D database table*
        """
        self.log.info('starting the ``generate_the_database_table`` method')

        tableName = self.dbTableName

        sqlQuery = u"""
            CREATE TABLE `%(tableName)s` (
              `primaryId` bigint(20) NOT NULL AUTO_INCREMENT COMMENT 'An internal counter',
              `Method` varchar(150) DEFAULT NULL,
              `dateCreated` datetime DEFAULT NULL,
              `dist_derived_from_sn` varchar(150) DEFAULT NULL,
              `dist_in_ned_flag` varchar(10) DEFAULT NULL,
              `dist_index_id` mediumint(9) DEFAULT NULL,
              `dist_mod` double DEFAULT NULL,
              `dist_mod_err` double DEFAULT NULL,
              `dist_mpc` double DEFAULT NULL,
              `galaxy_index_id` mediumint(9) DEFAULT NULL,
              `hubble_const` double DEFAULT NULL,
              `lmc_mod` double DEFAULT NULL,
              `notes` varchar(500) DEFAULT NULL,
              `primary_ned_id` varchar(150) DEFAULT NULL,
              `redshift` double DEFAULT NULL,
              `ref` varchar(150) DEFAULT NULL,
              `ref_date` int(11) DEFAULT NULL,
              `master_row` tinyint(4) DEFAULT '0',
              `major_diameter_arcmin` double DEFAULT NULL,
              `ned_notes` varchar(700) DEFAULT NULL,
              `object_type` varchar(100) DEFAULT NULL,
              `redshift_err` double DEFAULT NULL,
              `redshift_quality` varchar(100) DEFAULT NULL,
              `magnitude_filter` varchar(10) DEFAULT NULL,
              `minor_diameter_arcmin` double DEFAULT NULL,
              `morphology` varchar(50) DEFAULT NULL,
              `hierarchy` varchar(50) DEFAULT NULL,
              `galaxy_morphology` varchar(50) DEFAULT NULL,
              `radio_morphology` varchar(50) DEFAULT NULL,
              `activity_type` varchar(50) DEFAULT NULL,
              `in_ned` tinyint(4) DEFAULT NULL,
              `raDeg` double DEFAULT NULL,
              `decDeg` double DEFAULT NULL,
              `eb_v` double DEFAULT NULL,
              `sdss_coverage` TINYINT DEFAULT NULL,
              PRIMARY KEY (`primaryId`),
              UNIQUE KEY `galaxy_index_id_dist_index_id` (`galaxy_index_id`,`dist_index_id`)
            ) ENGINE=MyISAM AUTO_INCREMENT=0 DEFAULT CHARSET=latin1;
        """ % locals()
        try:
            dms.execute_mysql_write_query(
                sqlQuery=sqlQuery,
                dbConn=self.cataloguesDbConn,
                log=self.log
            )
        except:
            self.log.info(
                '`%(tableName)s` table already exists - continuing as normal' % locals())

        self.log.info('completed the ``generate_the_database_table`` method')
        return None

    def create_dictionary_of_nedd(
            self):
        """
        *create a list of dictionaries for the returned ned data*
        """
        self.log.info('starting the ``import_nedcsv_to_database`` method')

        count = 0
        with open(self.pathToDataFile, 'rb') as csvFile:
            csvReader = csv.reader(
                csvFile, dialect='excel', delimiter=',', quotechar='"')
            totalRows = sum(1 for row in csvReader)
        csvFile.close()
        totalCount = totalRows

        with open(self.pathToDataFile, 'rb') as csvFile:
            csvReader = csv.reader(
                csvFile, dialect='excel', delimiter=',', quotechar='"')
            theseKeys = []
            dictList = []
            for row in csvReader:
                if len(theseKeys) == 0:
                    totalRows -= 1
                if "Exclusion Code" in row and "Hubble const." in row:
                    for i in row:
                        if i == "redshift (z)":
                            theseKeys.append("redshift")
                        elif i == "Hubble const.":
                            theseKeys.append("hubble_const")
                        elif i == "G":
                            theseKeys.append("galaxy_index_id")
                        elif i == "err":
                            theseKeys.append("dist_mod_err")
                        elif i == "D (Mpc)":
                            theseKeys.append("dist_mpc")
                        elif i == "Date (Yr. - 1980)":
                            theseKeys.append("ref_date")
                        elif i == "REFCODE":
                            theseKeys.append("ref")
                        elif i == "Exclusion Code":
                            theseKeys.append("dist_in_ned_flag")
                        elif i == "Adopted LMC modulus":
                            theseKeys.append("lmc_mod")
                        elif i == "m-M":
                            theseKeys.append("dist_mod")
                        elif i == "Notes":
                            theseKeys.append("notes")
                        elif i == "SN ID":
                            theseKeys.append("dist_derived_from_sn")
                        elif i == "method":
                            theseKeys.append("dist_method")
                        elif i == "Galaxy ID":
                            theseKeys.append("primary_ned_id")
                        elif i == "D":
                            theseKeys.append("dist_index_id")
                        else:
                            theseKeys.append(i)
                    continue

                if len(theseKeys):
                    count += 1
                    if count > 1:
                        # Cursor up one line and clear line
                        sys.stdout.write("\x1b[1A\x1b[2K")
                    if count > totalCount:
                        count = totalCount
                    percent = (float(count) / float(totalCount)) * 100.
                    print "%(count)s / %(totalCount)s (%(percent)1.1f%%) rows added to memmory" % locals()
                    rowDict = {}
                    for t, r in zip(theseKeys, row):
                        rowDict[t] = r
                        if t == "ref_date":
                            try:
                                rowDict[t] = int(r) + 1980
                            except:
                                rowDict[t] = None

                    if rowDict["dist_index_id"] != "999999":
                        dictList.append(rowDict)

        csvFile.close()

        self.log.info('completed the ``import_nedcsv_to_database`` method')
        return dictList

    def clean_up_columns(
            self):
        """
        *clean up columns of the NED table*
        """
        self.log.info('starting the ``clean_up_columns`` method')

        tableName = self.dbTableName

        print "cleaning up %(tableName)s columns" % locals()

        sqlQuery = u"""
            update %(tableName)s set dist_mod_err = null where dist_mod_err = 0;
            update %(tableName)s set dist_in_ned_flag = null where dist_in_ned_flag = "";
            update %(tableName)s set notes = null where notes = "";
            update %(tableName)s set redshift = null where redshift = 0;
            update %(tableName)s set dist_derived_from_sn = null where dist_derived_from_sn = "";
            update %(tableName)s set hubble_const = null where hubble_const = 0;
            update %(tableName)s set lmc_mod = null where lmc_mod = 0;
            update %(tableName)s set master_row = 0;
            update %(tableName)s set master_row = 1 where primaryId in (select * from (select distinct primaryId from %(tableName)s group by galaxy_index_id) as alias);
        """ % locals()
        dms.execute_mysql_write_query(
            sqlQuery=sqlQuery,
            dbConn=self.cataloguesDbConn,
            log=self.log
        )

        self.log.info('completed the ``clean_up_columns`` method')
        return None

    def create_the_master_view(
            self):
        """
        *create the master view*
        """
        self.log.info('starting the ``create_the_master_view`` method')

        tableName = self.dbTableName

        print "creating view for %(tableName)s table" % locals()

        sqlQuery = u"""
            DROP VIEW IF EXISTS `view_%(tableName)s_master_recorders`;
        """ % locals()
        dms.execute_mysql_write_query(
            sqlQuery=sqlQuery,
            dbConn=self.cataloguesDbConn,
            log=self.log
        )
        sqlQuery = u"""
            CREATE
                VIEW `view_%(tableName)s_master_recorders` AS
                    (SELECT 
                        `%(tableName)s`.`primary_ned_id` AS `primary_ned_id`,
                        `%(tableName)s`.`object_type` AS `object_type`,
                        `%(tableName)s`.`raDeg` AS `raDeg`,
                        `%(tableName)s`.`decDeg` AS `decDeg`,
                        `%(tableName)s`.`dist_mpc` AS `dist_mpc`,
                        `%(tableName)s`.`dist_mod` AS `dist_mod`,
                        `%(tableName)s`.`dist_mod_err` AS `dist_mod_err`,
                        `%(tableName)s`.`Method` AS `dist_measurement_method`,
                        `%(tableName)s`.`redshift` AS `redshift`,
                        `%(tableName)s`.`redshift_err` AS `redshift_err`,
                        `%(tableName)s`.`redshift_quality` AS `redshift_quality`,
                        `%(tableName)s`.`major_diameter_arcmin` AS `major_diameter_arcmin`,
                        `%(tableName)s`.`minor_diameter_arcmin` AS `minor_diameter_arcmin`,
                        `%(tableName)s`.`magnitude_filter` AS `magnitude_filter`,
                        `%(tableName)s`.`eb_v` AS `gal_eb_v`,
                        `%(tableName)s`.`hierarchy` AS `hierarchy`,
                        `%(tableName)s`.`morphology` AS `morphology`,
                        `%(tableName)s`.`radio_morphology` AS `radio_morphology`,
                        `%(tableName)s`.`activity_type` AS `activity_type`,
                        `%(tableName)s`.`ned_notes` AS `ned_notes`,
                        `%(tableName)s`.`in_ned` AS `in_ned`,
                        `%(tableName)s`.`primaryId` AS `primaryId`
                    FROM
                        `%(tableName)s`
                    WHERE
                        (`%(tableName)s`.`master_row` = 1))
        """ % locals()
        dms.execute_mysql_write_query(
            sqlQuery=sqlQuery,
            dbConn=self.cataloguesDbConn,
            log=self.log
        )

        self.log.info('completed the ``create_the_master_view`` method')
        return None

    def get_metadata_for_galaxies(
            self):
        """
        *get metadata for galaxies*
        """
        self.log.info('starting the ``get_metadata_for_galaxies`` method')

        total, batches = self._count_galaxies_requiring_metadata()
        print "%(total)s galaxies require metadata. Need to send %(batches)s batch requests to NED." % locals()

        totalBatches = self.batches
        thisCount = 0

        # FOR EACH BATCH, GET THE GALAXY IDs, QUERY NED AND UPDATE THE DATABASE
        while self.total:
            thisCount += 1
            self._get_3000_galaxies_needing_metadata()
            self._query_ned_and_add_results_to_database(thisCount)
            self._count_galaxies_requiring_metadata()

        self.log.info('completed the ``get_metadata_for_galaxies`` method')
        return None

    def _count_galaxies_requiring_metadata(
            self):
        """
        *count galaxies requiring metadata*

        **Return:**
            - ``self.total``, ``self.batches`` -- total number of galaxies needing metadata & the number of batches required to be sent to NED
        """
        self.log.info(
            'starting the ``_count_galaxies_requiring_metadata`` method')

        tableName = self.dbTableName

        sqlQuery = u"""
            select count(*) as count from %(tableName)s where master_row = 1 and in_ned is null
        """ % locals()
        rows = dms.execute_mysql_read_query(
            sqlQuery=sqlQuery,
            dbConn=self.cataloguesDbConn,
            log=self.log
        )
        self.total = rows[0]["count"]
        self.batches = int(self.total / 3000.) + 1

        if self.total == 0:
            self.batches = 0

        self.log.info(
            'completed the ``_count_galaxies_requiring_metadata`` method')
        return self.total, self.batches

    def _get_3000_galaxies_needing_metadata(
            self):
        """
        *get 3000 galaxies needing metadata*

        **Return:**
            - ``len(self.theseIds)`` -- the number of NED IDs returned
        """
        self.log.info(
            'starting the ``_get_3000_galaxies_needing_metadata`` method')

        tableName = self.dbTableName

        # SELECT THE DATA FROM NED TABLE
        self.theseIds = []
        sqlQuery = u"""
            select primary_ned_id from %(tableName)s where master_row = 1 and in_ned is null limit 3000;
        """ % locals()
        rows = dms.execute_mysql_read_query(
            sqlQuery=sqlQuery,
            dbConn=self.cataloguesDbConn,
            log=self.log
        )
        for row in rows:
            self.theseIds.append(row["primary_ned_id"])

        self.log.info(
            'completed the ``_get_3000_galaxies_needing_metadata`` method')

        return len(self.theseIds)

    def _query_ned_and_add_results_to_database(
            self,
            batchCount):
        """
        *query ned and add results to database*

        **Key Arguments:**
            - ``batchCount`` - the index number of the batch sent to NED
        """
        self.log.info(
            'starting the ``_query_ned_and_add_results_to_database`` method')

        tableName = self.dbTableName

        # QUERY NED WITH BATCH
        totalCount = len(self.theseIds)
        print "requesting metadata from NED for %(totalCount)s galaxies (batch %(batchCount)s)" % locals()
        search = namesearch(
            log=self.log,
            names=self.theseIds,
            quiet=True
        )
        results = search.get()
        print "results returned from ned -- starting to add to database" % locals()

        # CLEAN THE RETURNED DATA AND UPDATE DATABASE
        totalCount = len(results)
        count = 0
        sqlQuery = ""
        for thisDict in results:
            thisDict["tableName"] = tableName
            count += 1
            for k, v in thisDict.iteritems():
                if not v or len(v) == 0:
                    thisDict[k] = "null"
                if k in ["major_diameter_arcmin", "minor_diameter_arcmin"] and (":" in v or "?" in v or "<" in v):
                    thisDict[k] = v.replace(":", "").replace(
                        "?", "").replace("<", "")
                if isinstance(v, str) and '"' in v:
                    thisDict[k] = v.replace('"', '\\"')
            if "Input name not" not in thisDict["input_note"] and "Same object as" not in thisDict["input_note"]:
                thisDict[
                    "raDeg"] = dat.ra_sexegesimal_to_decimal.ra_sexegesimal_to_decimal(thisDict["ra"])
                thisDict[
                    "decDeg"] = dat.declination_sexegesimal_to_decimal.declination_sexegesimal_to_decimal(thisDict["dec"])

                sqlQuery += u"""
                    update %(tableName)s
                        set redshift_quality = "%(redshift_quality)s",
                        redshift = %(redshift)s,
                        hierarchy = "%(hierarchy)s",
                        object_type = "%(object_type)s",
                        major_diameter_arcmin = %(major_diameter_arcmin)s,
                        morphology = "%(morphology)s",
                        magnitude_filter = "%(magnitude_filter)s",
                        ned_notes = "%(ned_notes)s",
                        eb_v = %(eb-v)s,
                        raDeg = %(raDeg)s,
                        radio_morphology = "%(radio_morphology)s",
                        activity_type = "%(activity_type)s",
                        minor_diameter_arcmin = %(minor_diameter_arcmin)s,
                        decDeg = %(decDeg)s,
                        redshift_err = %(redshift_err)s,
                        in_ned = 1
                    where primary_ned_id = "%(input_name)s" and master_row = 1;\n
                """ % thisDict
            else:
                sqlQuery += u"""
                    update %(tableName)s
                        set in_ned = 0
                    where primary_ned_id = "%(input_name)s" and master_row = 1;\n
                """ % thisDict
                sqlQuery = sqlQuery.replace('"null"', 'null')
        sqlQuery = sqlQuery.replace('"null"', 'null')

        rows = dms.execute_mysql_read_query(
            sqlQuery=sqlQuery,
            dbConn=self.cataloguesDbConn,
            log=self.log
        )
        print "%(count)s/%(totalCount)s galaxy metadata batch entries added to database" % locals()
        if count < totalCount:
            # Cursor up one line and clear line
            sys.stdout.write("\x1b[1A\x1b[2K")

        self.log.info(
            'completed the ``_query_ned_and_add_results_to_database`` method')
        return None

    def _update_sdss_coverage(
            self):
        """
        *update sdss coverage*
        """
        self.log.info('starting the ``_update_sdss_coverage`` method')

        tableName = self.dbTableName

        # SELECT THE LOCATIONS NEEDING TO BE CHECKED
        sqlQuery = u"""
            select primary_ned_id, primaryID, raDeg, decDeg, sdss_coverage from %(tableName)s where sdss_coverage is null and master_row = 1 and in_ned = 1 order by dist_mpc;
        """ % locals()
        rows = dms.execute_mysql_read_query(
            sqlQuery=sqlQuery,
            dbConn=self.cataloguesDbConn,
            log=self.log
        )

        totalCount = len(rows)
        count = 0
        for row in rows:
            count += 1
            if count > 1:
                # Cursor up three lines and clear
                sys.stdout.write("\x1b[1A\x1b[2K")
                sys.stdout.write("\x1b[1A\x1b[2K")
                sys.stdout.write("\x1b[1A\x1b[2K")

            if count > totalCount:
                count = totalCount
            percent = (float(count) / float(totalCount)) * 100.

            primaryID = row["primaryID"]
            raDeg = float(row["raDeg"])
            decDeg = float(row["decDeg"])
            primary_ned_id = row["primary_ned_id"]

            # SDSS CAN ONLY ACCEPT 60 QUERIES/MIN
            time.sleep(1.1)
            print "%(count)s / %(totalCount)s (%(percent)1.1f%%) NED galaxies checked for SDSS coverage" % locals()
            print "NED NAME: ", primary_ned_id
            sdss_coverage = dat.check_for_sdss_coverage.check_for_sdss_coverage(
                log=self.log,
                raDeg=raDeg,
                decDeg=decDeg
            )
            if sdss_coverage == 999:
                sdss_coverage_flag = "null"
            elif sdss_coverage == True:
                sdss_coverage_flag = 1
            elif sdss_coverage == False:
                sdss_coverage_flag = 0
            else:
                self.log.error('cound not get sdss coverage' % locals())
                sys.exit(0)

            # UPDATE THE DATABASE FLAG
            sqlQuery = u"""
                update %(tableName)s set sdss_coverage = %(sdss_coverage_flag)s where primaryID = %(primaryID)s
            """ % locals()
            dms.execute_mysql_write_query(
                sqlQuery=sqlQuery,
                dbConn=self.cataloguesDbConn,
                log=self.log
            )

        self.log.info('completed the ``_update_sdss_coverage`` method')
        return None

    # use the tab-trigger below for new method
    # xt-class-method



import os
import nose
import shutil
import yaml
from sherlock import crossmatcher
from sherlock import database
from sherlock.utKit import utKit

# load settings
stream = file(
    "/Users/Dave/git_repos/sherlock/example_settings.yaml", 'r')
settings = yaml.load(stream)
stream.close()


# SETUP AND TEARDOWN FIXTURE FUNCTIONS FOR THE ENTIRE MODULE
moduleDirectory = os.path.dirname(__file__)
utKit = utKit(moduleDirectory)
log, dbConn, pathToInputDir, pathToOutputDir = utKit.setupModule()
utKit.tearDownModule()

# xnose-class-to-test-main-command-line-function-of-module


class test_crossmatcher():

    def test_crossmatcher_function(self):
        kwargs = {}
        kwargs["log"] = log
        kwargs["settings"] = settings
        # xt-kwarg_key_and_value
        testObject = database(**kwargs)
        transientsDbConn, cataloguesDbConn = testObject.get()

        kwargs = {}
        kwargs["log"] = log
        kwargs["dbConn"] = cataloguesDbConn
        kwargs["settings"] = settings
        kwargs["transients"] = [{'ps1_designation': u'PS1-14aef', 'name': u'4L3Piiq', 'detection_list_id': 2, 'local_comments': u'', 'ra': 0.02548233704918263, 'followup_id': 2065412L, 'dec': -4.284933417540423, 'id': 1000006110041705700L, 'object_classification': 0L}, {'ps1_designation': u'PS1-13dcr', 'name': u'3I3Phzx', 'detection_list_id': 2, 'local_comments': u'', 'ra': 4.754236999477372, 'followup_id': 1140386L, 'dec': 28.276703631398625, 'id': 1001901011281636100L, 'object_classification': 0L}, {'ps1_designation': u'PS1-13dhc', 'name': u'3I3Pixd', 'detection_list_id': 2, 'local_comments': u'', 'ra': 1.3324973428505413, 'followup_id': 1202386L, 'dec': 32.98869220595689, 'id': 1000519791325919200L, 'object_classification': 0L}]
        # xt-kwarg_key_and_value

        testObject = crossmatcher(**kwargs)
        testObject.get()

        # x-print-testpage-for-pessto-marshall-web-object

    # x-class-to-test-named-worker-function
